from pathlib import Path
from typing import Iterator, List, Dict, Any, Tuple, Optional
from pptx import Presentation
from pptx.util import Inches, Pt
from pptx.enum.shapes import MSO_SHAPE
from pptx.dml.color import RGBColor


class PowerPointHandler:
    def create_empty_presentation(self, file_path: Path):
        """
        Create an empty PowerPoint presentation.

        :param file_path: Path where the new PowerPoint file will be saved.
        """
        prs = Presentation()
        prs.save(file_path)

    def create_presentation_from_text_list(self, text_list: List[str], file_path: Path):
        """
        Create a PowerPoint file where each slide contains text from a list.

        :param text_list: List of texts to add to slides.
        :param file_path: Path where the new PowerPoint file will be saved.
        """
        prs = Presentation()
        for text in text_list:
            slide = prs.slides.add_slide(prs.slide_layouts[1])  # Title and Content layout
            textbox = slide.shapes.placeholders[1]
            textbox.text = text
        prs.save(file_path)

    def create_presentation_from_template(
        self, template_path: Path, slide_contents: Iterator[Dict[str, Any]], file_path: Path
    ):
        """
        Create a presentation from a template, adding texts or images per slide.

        :param template_path: Path to the PowerPoint template file.
        :param slide_contents: An iterator where each item is a dict with content for a slide.
        :param file_path: Path where the new PowerPoint file will be saved.
        """
        prs = Presentation(template_path)
        for content in slide_contents:
            slide = prs.slides.add_slide(prs.slide_layouts[1])  # Adjust layout as needed
            self._add_content_to_slide(slide, content)
        prs.save(file_path)

    def append_slides_to_presentation(
        self, file_path: Path, slide_contents: Iterator[Dict[str, Any]]
    ):
        """
        Edit a PowerPoint file to append more slides from an iterator.

        :param file_path: Path to the existing PowerPoint file.
        :param slide_contents: An iterator where each item is a dict with content for a slide.
        """
        prs = Presentation(file_path)
        for content in slide_contents:
            slide = prs.slides.add_slide(prs.slide_layouts[1])  # Adjust layout as needed
            if 'text' in content:
                textbox = slide.shapes.placeholders[1]
                textbox.text = content['text']
            if 'image' in content:
                left = Inches(1)
                top = Inches(2)
                slide.shapes.add_picture(content['image'], left, top, width=Inches(5))
        prs.save(file_path)

    def create_slide(
        self,
        text_content: Optional[str] = None,
        image_path: Optional[str] = None,
        image_size: Optional[Tuple[float, float, float]] = None,
        font_size: float = 18,
        font_color: str = '000000',
        font_name: str = 'Calibri'
    ) -> Dict[str, Any]:
        """
        Create a slide content dictionary.

        :param text_content: Text to include in the slide.
        :param image_path: Path to the image file to include in the slide.
        :param image_size: Tuple of three elements (left, top, width) for image positioning.
        :param font_size: Font size for the text.
        :param font_color: Font color for the text in hex format (e.g., 'FF0000' for red).
        :param font_name: Font name for the text.
        :return: A dictionary representing the slide content.
        """
        slide_content = {}

        if text_content:
            slide_content['text'] = {
                'content': text_content,
                'font_size': Pt(font_size),
                'font_color': font_color,
                'font_name': font_name
            }

        if image_path and image_size:
            left, top, width = image_size
            slide_content['image'] = {
                'path': image_path,
                'left': Inches(left),
                'top': Inches(top),
                'width': Inches(width)
            }

        return slide_content

    def _add_content_to_slide(self, slide, content: Dict[str, Any]):
        from pptx.dml.color import RGBColor

        # Handle text content
        if 'text' in content:
            text_info = content['text']
            # Always add a textbox
            left = Inches(1)
            top = Inches(1)
            width = Inches(8)
            height = Inches(2)
            textbox = slide.shapes.add_textbox(left, top, width, height)
            text_frame = textbox.text_frame

            p = text_frame.paragraphs[0]
            run = p.add_run()
            run.text = text_info.get('content', '')
            font = run.font
            font.name = text_info.get('font_name', 'Calibri')
            font.size = text_info.get('font_size', Pt(18))
            font.color.rgb = RGBColor.from_string(text_info.get('font_color', '000000'))

        # Handle image content (unchanged)
        if 'image' in content:
            image_info = content['image']
            img_path = image_info['path']
            left = image_info.get('left', Inches(1))
            top = image_info.get('top', Inches(2))
            width = image_info.get('width', None)
            height = image_info.get('height', None)
            slide.shapes.add_picture(img_path, left, top, width=width, height=height)
